#undef OS2Change_Keyboard
#define min(a,b) ((a)<(b))?(a):(b)
#define max(a,b) ((a)>(b))?(a):(b)
#define cap(val,lo,up) max(lo,min(val,up))

// Syscalls
#define SCA 0xD201D002
#define SCB 0x422B0009
#define SCE 0x80010070
typedef int(*sc_iv)(void);
const unsigned int sc003B[] = { SCA, SCB, SCE, 0x003B };
#define RTC_GetTicks (*(sc_iv)sc003B)

extern "C"
{
	#include <fxlib.h>
	#include "sh4_compatibility.h"
}

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stdio.h>

const int dirMappings[8][2] = {
	{0,-1},
	{1,-1},
	{1, 0},
	{1, 1},
	{0, 1},
	{-1,1},
	{-1,0},
	{-1,-1}
};

void render(char board[6][7], int pX)
{
	unsigned char empty[3] = {0xE6, 0xA3, 0};
	unsigned char player[3] = {0xE6, 0xD3, 0};
	unsigned char comput[3] = {0xE6, 0xA4, 0};
	unsigned char cursor[3] = {0xE6, 0x9D, 0};
	unsigned char arrowLeft[3] = {0xE6, 0x90, 0};
	unsigned char arrowRight[3] = {0xE6, 0x91, 0};
	unsigned char arrowDown[3] = {0xE6, 0x93, 0};
	Bdisp_AllClr_VRAM();
	locate(pX + 1, 1);
	Print(cursor);
	for(int y = 0; y < 6; y++)
	{
		for(int x = 0; x < 7; x++)
		{
			locate(x + 1, y + 2);
			switch(board[y][x])
			{
				case 0:
					Print(empty);
					break;
				case -1:
					Print(player);
					break;
				case 1:
					Print(comput);
					break;
			}
		}
	}
	locate(9, 6);
	Print(arrowLeft);
	Print(arrowRight);
	Print(" move");
	locate(9, 7);
	Print(arrowDown);
	Print(" drop");
	Bdisp_PutDisp_DD();
}

void win(char board[6][7], int pX, int whom)
{
	unsigned int key;
	const DISPBOX rect = {45, 40, 127, 63};

	render(board, pX);
	if(whom == -1) PrintXY(64, 28, "YOU WIN!", 0);
	if(whom == 1) PrintXY(64, 28, "I WIN!", 0);
	if(whom == 0) PrintXY(64, 28, "DRAW!", 0);
	Bdisp_AreaClr_VRAM(&rect);
	locate(9, 7);
	Print("[EXE] restart");
	Bdisp_PutDisp_DD();
	memset(board, 0, 42 * sizeof(int));
	while(1)
	{
		GetKey(&key);
		if(key == KEY_CTRL_EXE) return;
	}
}

bool checkIfFull(char board[6][7])
{
	int isFull = true;

	for(int y = 0; y < 6; y++)
	{
		for(int x = 0; x < 7; x++)
		{
			if(board[y][x] == 0)
			{
				isFull = false;
				break;
			}
		}
		if(!isFull) break;
	}
	return isFull;
}

int checkWin(char board[6][7])
{
	int checking;
	int endX, endY;
	int dX, dY;

	for(int y = 0; y < 6; y++)
	{
		for(int x = 0; x < 7; x++)
		{
			if(board[y][x] == 0) continue;
			for(int dir = 0; dir < 8; dir++)
			{
				dX = dirMappings[dir][0];
				dY = dirMappings[dir][1];
				endX = x + dX * 3;
				endY = y + dY * 3;
				if(endX < 0 || endX > 6 || endY < 0 || endY > 5) continue;
				checking = board[y][x];
				if(board[y + dY][x + dX] != checking) continue;
				if(board[y + (dY << 1)][x + (dX << 1)] != checking) continue;
				if(board[endY][endX] != checking) continue;
				return checking;
			}
		}
	}
	return 0;
}

int getValidY(char board[6][7], int x)
{
	if(board[0][x] != 0) return -1;
	for(int y = 1; y < 6; y++)
	{
		if(board[y][x] != 0) return y - 1;
	}
	return 5;
}

int minmax(char board[6][7], int depth, bool maximise)
{
	int value;
	int bestValue;
	int y;

	value = checkWin(board);
	if(value != 0)
	{
		if(value == -1) return -10 + depth;
		if(value == 1) return 10 - depth;
	}

	if(checkIfFull(board)) return 0;

	if(depth == 5) return 0;

	if(maximise) bestValue = -1000;
	if(!maximise) bestValue = 1000;

	for(int x = 0; x < 7; x++)
	{
		y = getValidY(board, x);
		if(y == -1) continue;
		if(maximise) board[y][x] = 1;
		if(!maximise) board[y][x] = -1;
		value = minmax(board, depth + 1, !maximise);
		board[y][x] = 0;
		if(maximise) bestValue = max(bestValue, value);
		if(!maximise) bestValue = min(bestValue, value);
	}
	return bestValue;
}

void addCounter(char board[6][7], int x, int whom, int pX)
{
	int end = getValidY(board, x);
	for(int y = 0; y <= end; y++)
	{
		board[y][x] = whom;
		if(y > 0) board[y - 1][x] = 0;
		render(board, pX);
		Sleep(100);
	}
}

int main()
{
	char board[6][7];
	memset(board, 0, 42 * sizeof(int));
	int value;
	int bestValue;
	unsigned int key;
	int pX = 0;
	int y;
	char choices[7];
	int last;
	int selected;

	srand(RTC_GetTicks());

	while(1)
	{
		render(board, pX);
		while(1)
		{
			GetKey(&key);
			if(key == KEY_CTRL_LEFT) pX--;
			if(key == KEY_CTRL_RIGHT) pX++;
			pX = cap(pX, 0, 6);
			render(board, pX);

			if(key == KEY_CTRL_DOWN)
			{
				y = getValidY(board, pX);
				if(y != -1)
				{
					addCounter(board, pX, -1, pX);
					break;
				}
			}
		}

		render(board, pX);
		if(checkWin(board))
		{
			win(board, pX, checkWin(board));
			pX = 0;
			continue;
		}
		if(checkIfFull(board))
		{
			win(board, pX, 0);
			pX = 0;
			continue;
		}

		PrintXY(52, 28, "Thinking...", 0);
		Bdisp_PutDisp_DD();
		bestValue = -1000;
		last = -1;
		memset(choices, -1, 7 * sizeof(char));
		for(int x = 0; x < 7; x++)
		{
			y = getValidY(board, x);
			if(y == -1) continue;
			board[y][x] = 1;
			value = minmax(board, 1, false);
			board[y][x] = 0;
			if(value > bestValue)
			{
				memset(choices, -1, 7 * sizeof(char));
				choices[0] = x;
				last = 0;
				bestValue = value;
			}
			else if(value == bestValue)
			{
				last++;
				choices[last] = x;
			}
		}

		selected = rand() % last;
		addCounter(board, choices[selected], 1, pX);

		render(board, pX);
		if(checkWin(board))
		{
			win(board, pX, checkWin(board));
			pX = 0;
			continue;
		}
		if(checkIfFull(board))
		{
			win(board, pX, 0);
			pX = 0;
			continue;
		}
	}
    return 1;
}

extern "C"
{
	int AddIn_main(int isAppli, unsigned short OptionNum)
	{return main();}
	#pragma section _BR_Size
	unsigned long BR_Size;
	#pragma section
	#pragma section _TOP
	int InitializeSystem(int isAppli, unsigned short OptionNum)
	{return INIT_ADDIN_APPLICATION(isAppli, OptionNum);}
	#pragma section
}